'use strict';

const request = require('request-promise'),
  _ = require('lodash'),
  logger = require('../logger'),
  util = require('util');

class SpectrumConnector {

  /**
   *
   * @param provider event emitter that will provide data for handling
   * @param options all other required options
   */
  constructor(provider, options) {
    this._provider = provider;
    this._options = options;
    this._requestOptions = {};
  }

  initialize() {
    let authHash = Buffer.from(this._options.user.name + ':' + this._options.user.password).toString('base64');
    let uri = this._options.uri
        ? this._options.uri
        : `${this._options.protocol}://${this._options.host}:${this._options.port}${this._options.resource}`;
    this._requestOptions = {
      headers: {
        'Authorization': 'Basic ' + authHash
      },
      method: 'GET',
      json: true,
      uri: uri,
      timeout: parseInt(this._options.timeoutSec) * 1000
    };
    this._provider.on('event', async (data) => {
      logger.info(`Received event from panel: ${data.panelId}`);
      await this._sendRequest(data);
    })
  }

  finalize() {
    this._provider.removeAllListeners('event');
    this._requestOptions = {};
  }

  async _sendRequest(data) {
    this._requestOptions.qs = this._parseParameters(data.event);

    try {
      let response = await request(this._requestOptions);
      logger.info('Received a response:', response);
    } catch (err) {
      logger.error('API call failed', err);
    }
  }

  _parseParameters(liveEvent) {
    let doorPlaceholder = _.find(liveEvent.placeholders, ['type', 'door']);
    // Placeholders are objects, we need only 'text' property to be placed into info string
    let placeholders = _.map(liveEvent.placeholders, (placeholder) => placeholder.text);
    let parameters = {
      timestamp: liveEvent.occurred,
      source: doorPlaceholder ? doorPlaceholder.text : null,
      caption: liveEvent.topic,
      description: _.isEmpty(placeholders) ? liveEvent.info : util.format(liveEvent.info, ...placeholders)
    };
    logger.info('Request parameters:', parameters);
    return parameters;
  }
}

module.exports = SpectrumConnector;