'use strict';

const EventEmitter = require('events').EventEmitter,
    pdkClient = require('@pdk/client'),
    utils = require('../utils/server-utils'),
    logger = require('../logger'),
    InvalidTokenError = require('../errors/InvalidTokenError'),
    _ = require('lodash');

class CloudNodeDataProvider extends EventEmitter {
  /**
   * Class constructor. Store all incoming parameters without any initialization.
   * @param options configuration options required for the pdk-client
   */
  constructor(authenticator, options) {
    super(options);

    this._authenticator = authenticator;
    this._options = options;
    this._streams = {};
    this._reconnectTimers = {};
  }

  /**
   * Initialize pdk-client, subscribe on events coming via Web socket
   * @return {Promise<void>}
   */
  async initialize() {
    try {
      const authSession = await pdkClient.makeSession(this._authenticator, this._options.issuerApiUri);

      let panelIds = this._parsePanelIds(this._options.panelId);

      await this._connectToPanels(panelIds, authSession);
    } catch (err) {
      if (err.message.includes('invalid_grant')) {
        throw new InvalidTokenError();
      }
      throw err;
    }
  }

  finalize() {
    logger.debug('Finalize cloud node data provider.');
    _.forEach(this._streams, (stream) => {
      stream.close();
    });
    this._streams = {};

    _.forEach(this._reconnectTimers, (timer) => {
      clearTimeout(timer);
    })
  }

  async _connectToPanels(panelIds, authSession) {
    for (let panelId of panelIds) {
      try {
        let panel = await authSession('panels/' + panelId);

        await this._connectSocket(panel, authSession, panelId);
      } catch(err) {
        logger.error(`Failed connect to panel: ${panelId}, Error: ${err}`);
      }
    }
  }

  _startReconnectTimer(panel, authSession, panelId, stream) {
    stream.close();
    this._stopReconnectTimer(panelId);

    logger.debug(`Start timer to reconnect to panel ${panelId}`);
    this._reconnectTimers[panelId] = setTimeout(() => this._connectSocket(panel, authSession, panelId),
      parseInt(this._options.reconnectTimeoutSec) * 1000);
  }

  _stopReconnectTimer(panelId) {
    if (this._reconnectTimers[panelId]) {
      logger.debug(`Clear reconnect timer for panel ${panelId}`);
      clearTimeout(this._reconnectTimers[panelId]);
      delete this._reconnectTimers[panelId];
    }
  }

  async _connectSocket(panel, authSession, panelId) {
    try {
      logger.debug(`Start connecting to ${panelId}`);
      // Create an authentication session to the panel's API
      const panelSession = await pdkClient.makePanelSession(authSession, panel);

      let stream = panelSession.createEventStream();
      this._streams[panelId] = stream;

      stream.on('error', (error) => {
        logger.error('Error happened:', error);

        this._startReconnectTimer(panel, authSession, panelId, stream);
      });

      stream.on('liveEvent', (arrivedEvent) => {
        let data = {
          event: arrivedEvent,
          panelId: panelId
        };
        this.emit('event', data);
      });

      stream.on('connect', () => {
        logger.info(`Connection to panel ${panelId} has been established`);
        this._stopReconnectTimer(panelId);
      });

      stream.on('disconnect', (reason) => {
        logger.info(`Connection to panel ${panelId} has been lost. ${reason}`);
        // the disconnection was initiated by the server, we need to reconnect manually in this case
        if (reason === 'io server disconnect') {
          this._startReconnectTimer(panel, authSession, panelId, stream);
        }
      });

      stream.on('connect_error', (error) => {
        logger.error(`An error happened during connection to the panel ${panelId}.`, error);
      });

      stream.on('connect_timeout', () => {
        logger.error(`Connection to panel ${panelId} timed out.`);
      });

      stream.on('serverError', (error) => {
        logger.error(`serverError event has been received from panel ${panelId}.`, error);
      });
    } catch (err) {
      if (err.message.includes('invalid_grant')) {
        logger.error(new InvalidTokenError());
        process.exit(1);
        return;
      }
      logger.error(err);
    }
  }

  sendCommand(options) {
    let stream = this._streams[options.panelId];

    if(_.isNil(stream) || !stream.connected) {
      throw new Error(`No connected socket for panel: ${options.panelId}`);
    }

    for(let doorId of options.doorIds) {
      let command = utils.populateCommand(doorId, options.command);
      stream.emit('command', command);
      logger.info(`Send command: ${options.command} to door: ${doorId} on panel: ${options.panelId}`);
    }
  }

  _parsePanelIds(panelId) {
    if(_.isArray(panelId)) {
      return panelId;
    }
    return _.split(panelId, ';');
  }
}

module.exports = CloudNodeDataProvider;