'use strict';

const config = require('./configuration/index');
const logger = require('./logger');
const DataProvider = require('./data-providers/index').CloudNodeDataProvider;
const pdkClient = require('@pdk/client');
const connectors = require('./connectors/index');
const FileAdapter = require('./adapters/file-adapter');
const commandServerUtils = require('./server/command-server');
const authenticatorUtils = require('./utils/authenticator');

let dataProvider, connector, commandServer;

async function start() {

  // Get operational mode
  let mode = config.mode;

  switch (mode) {
    case 'setup':
      return await _startSetupMode();
    case 'setup-print-url':
      return await _startSetupMode(false);
    case 'spectrum':
      return await _startSpectrumMode();
    default:
      throw new Error('Unexpected mode');
  }

}

async function _startSetupMode(openBrowser = true) {
  let authParams = {
    client_id: config.pdkClient.clientId,
    client_secret: config.pdkClient.clientSecret,
    issuer: config.pdkClient.issuerUri,
    // specify offline_access scope to retrieve refresh token from the issuer
    scope: 'openid offline_access'
  };

  // Print url if we don't need open a browser automatically
  if (!openBrowser) {
    authParams.opener = console.log;
  }

  // create authenticator instance with user flow to retrieve refresh token
  let userAuthenticator = pdkClient.userauth(authParams);

  // authenticator returns function that can be used to get refresh token
  let tokenSetFunc = await userAuthenticator();

  // get token set
  let tokenSet = await tokenSetFunc();

  let fileAdapter = new FileAdapter({
    path: config.pdkClient.refreshTokenPath
  });

  await fileAdapter.setToken(tokenSet.refresh_token);
}

async function _startSpectrumMode() {

  let fileAdapter = new FileAdapter({
    path: config.pdkClient.refreshTokenPath
  });

  let token = await fileAdapter.getToken();

  if (!token) {
    throw new Error('Unable to find refresh token locally. Please run the tool with --mode=setup flag to get a token.');
  }

  let userAuth = pdkClient.userauth({
    client_id: config.pdkClient.clientId,
    client_secret: config.pdkClient.clientSecret,
    issuer: config.pdkClient.issuerUri,
    refresh_token: token
  });

  // Wrap standard user authentication to additionally store new refresh tokens in the file system
  let userAuthWrapped = authenticatorUtils.wrapUserAuthenticator(userAuth, fileAdapter);

  dataProvider = new DataProvider(userAuthWrapped, {
    issuerApiUri: config.pdkClient.issuerUri + '/api/',
    panelId: config.pdkClient.panelId,
    reconnectTimeoutSec: config.pdkClient.reconnectTimeoutSec
  });

  await dataProvider.initialize();

  // Start server for handling commands
  commandServer = commandServerUtils.startServer(dataProvider, config.commandServer.port);

  //TODO: check configuration options and choose needed connector if we want to support multiple destinations
  connector = new connectors.SpectrumConnector(dataProvider, {
    uri: config.spectrum.uri,
    protocol: config.spectrum.protocol,
    host: config.spectrum.host,
    port: config.spectrum.port,
    resource: config.spectrum.resource,
    timeoutSec: config.spectrum.request.timeoutSec,
    user: {
      name: config.spectrum.user.name,
      password: config.spectrum.user.password
    }
  });

  await connector.initialize();
  logger.info('Started listening for cloud nodes events.');
}

function stop() {
  if (dataProvider) {
    dataProvider.finalize();
  }
  commandServer.close();
  connector.finalize();
}

start()
  .catch((err) => {
    logger.error(err);
  });

process.on('SIGTERM', stop);
process.on('SIGINT', stop);

module.exports = {
  start,
  stop
};
